;; Generated package description from xr.el  -*- no-byte-compile: t -*-
(define-package "xr" "2.2.0.20260205.105100" "Convert string regexp to rx notation" '((emacs "27.1")) :commit "fc4d6b5ddae9eb2c8adc2b117e6f33f3a09a4e96" :authors '(("Mattias Engdegård" . "mattias.engdegard@gmail.com")) :maintainer '("Mattias Engdegård" . "mattias.engdegard@gmail.com") :keywords '("lisp" "regexps") :url "https://github.com/mattiase/xr")
